
#include <sys/p2es_clock.h>
#include <inttypes.h>
#include <stdio.h>
#include <stdlib.h>
#include <math.h>


#define MIN(a,b) (((a)<(b))?(a):(b))
#define MAX(a,b) (((a)>(b))?(a):(b))



enum {
	_clkfreq = 10_000_000,			// system clock frequency

	SAMPIOD = 500,				// decimation/sample period, in sysclocks
	ADCMODE = P_ADC_1X | 3,			// ADC source/gain, and Sinc order, can include P_OE for overlaying a preset square wave on the recording

	ADCPIN = 40,				// base pin number for the eight ADC inputs
	ADCPINS = ADCPIN | (7<<6),
	SAMPLES = 10000,			// total samples recorded per channel
};



//============================================================================

__pasm {
ORG 0
startADC
		setq	#3
		rdlong	.launcher, ptra			' fetch the stored cog ID of launcher cog ... and the other three parameters
		rczr	.launcher+1		wcz	' extract Sinc# parameter
		shl	.launcher+1, #2			' restore pin mode
	if_nc	or	.launcher+1, #P_COUNT_HIGHS	' Sinc1
	if_c	or	.launcher+1, #P_ADC		' Sinc2 & 3

		fltl	#ADCPINS
		wrpin	.launcher+1, #ADCPINS		' turn on ADC modulators and input/gain, and select smartpin mode
	if_nc	wxpin	.launcher+2, #ADCPINS		' set Sinc1 decimation/sample period in sysclocks
	if_nc	wypin	#0, #ADCPINS
if_c_and_nz	wxpin	#%01_0000, #ADCPINS		' Sinc2 filter sub-mode
if_c_and_z	wxpin	#%10_0000, #ADCPINS		' Sinc3 filter sub-mode
	if_c	wypin	.launcher+2, #ADCPINS		' set decimation period in sysclocks
		dirh	#ADCPINS			' enable smartpins
		waitx	##_clkfreq/100			' 10 ms pause for ADC modulators to warm up

		wrfast	.launcher+3, ptra		' initiate FIFO with specified wrap length, PTRA is hubRAM address of "mailbox"
		rdpin	pa, #ADCPIN			' re-arm smartpin, make IN low
		setse1	#(%001<<6) | ADCPIN		' event trigger on rising edge of IN
		waitse1					' start synchronised
if_c_and_z	jmp	#.sinc3

.loop1
		mov	pb, #0				' channel index

		rep	@.chrend, #8			' eight channels

		alts	pb, #ADCPIN			' non-incrementing index [ADCPIN..ADCPIN+7]
		rdpin	pa, #0-0			' read decimation data from smartpin

	if_c	alts	pb, #.diff2			' non-incrementing index [diff2..diff2+7]
	if_c	sub	pa, 0-0				' differentiate (difference with prior decimation)

		altd	pb, .altinc			' incrementing index [diff2..diff2+7]
	if_c	add	0-0, pa				' store decimation (adjust prior data by difference)

	if_c	zerox	pa, #26				' truncate to match 27-bit hardware accumulators

		wflong	pa				' write to hubRAM
.chrend

		waitse1					' wait for next decimation event
		jnfbw	#.loop1				' loop until FIFO wrapping event, don't want remaining FIFO data
		jmp	#.done

.sinc3
		rdpin	pa, #ADCPIN
		sub	pa, .diff1
		add	.diff1, pa
		sub	pa, .diff2
		add	.diff2, pa
		zerox	pa, #26
		wflong	pa

		rdpin	pa, #ADCPIN+1
		sub	pa, .diff1+1
		add	.diff1+1, pa
		sub	pa, .diff2+1
		add	.diff2+1, pa
		zerox	pa, #26
		wflong	pa

		rdpin	pa, #ADCPIN+2
		sub	pa, .diff1+2
		add	.diff1+2, pa
		sub	pa, .diff2+2
		add	.diff2+2, pa
		zerox	pa, #26
		wflong	pa

		rdpin	pa, #ADCPIN+3
		sub	pa, .diff1+3
		add	.diff1+3, pa
		sub	pa, .diff2+3
		add	.diff2+3, pa
		zerox	pa, #26
		wflong	pa

		rdpin	pa, #ADCPIN+4
		sub	pa, .diff1+4
		add	.diff1+4, pa
		sub	pa, .diff2+4
		add	.diff2+4, pa
		zerox	pa, #26
		wflong	pa

		rdpin	pa, #ADCPIN+5
		sub	pa, .diff1+5
		add	.diff1+5, pa
		sub	pa, .diff2+5
		add	.diff2+5, pa
		zerox	pa, #26
		wflong	pa

		rdpin	pa, #ADCPIN+6
		sub	pa, .diff1+6
		add	.diff1+6, pa
		sub	pa, .diff2+6
		add	.diff2+6, pa
		zerox	pa, #26
		wflong	pa

		rdpin	pa, #ADCPIN+7
		sub	pa, .diff1+7
		add	.diff1+7, pa
		sub	pa, .diff2+7
		add	.diff2+7, pa
		zerox	pa, #26
		wflong	pa

		waitse1					' wait for next decimation event
		jnfbw	#.sinc3				' loop until FIFO wrapping event, don't want remaining FIFO data

.done
		dirl	#ADCPINS
		wrpin	#0, #ADCPINS
		decod	.launcher
		cogatn	.launcher
		cogid	pa
		cogstop	pa


.altinc		long	.diff2 | (1<<9)
.launcher	long	0
.diff1		res	8
.diff2		res	8
}



static void  emit_adc_mode( uint32_t mode )
{
	printf( " Sinc order: %d\n ADC source/gain: ", mode&3 );
	mode = (mode >> 15) & 7;
	switch( mode )
	{
		case 0:
			puts( "P_ADC_GIO" );
			break;
		case 1:
			puts( "P_ADC_VIO" );
			break;
		case 2:
			puts( "P_ADC_FLOAT" );
			break;
		case 3:
			puts( "P_ADC_1X" );
			break;
		case 4:
			puts( "P_ADC_3X" );
			break;
		case 5:
			puts( "P_ADC_10X" );
			break;
		case 6:
			puts( "P_ADC_30X" );
			break;
		case 7:
			puts( "P_ADC_100X" );
			break;
		default:
			puts( "unknown" );
			break;
	}	
}



int  main( void )
{
	int32_t volatile * mailbox;
	int32_t  idx;


	mailbox = __builtin_alloca( SAMPLES * 32 + 8 );		// 8 channels x 4 bytes (32-bit samples)
	if( mailbox == NULL )  return( 1 );
	memset( mailbox, 0, SAMPLES * 32 + 8 );

	mailbox[0] = _cogid();					// launcher (self) cog ID for launchee COGATN
	mailbox[1] = ADCMODE;					// ADC source/gain, and Sinc order
	mailbox[2] = SAMPIOD;					// decimation/sample period, in sysclocks
	mailbox[3] = SAMPLES/2;					// buffer length, measured in 64-byte FIFO blocks

	_waitms( 200 );						// pause for terminal ready
	printf( "\n\n\n\n\n\n\n\n\n\n Sysclock frequency: %d,%03d,%03d Hz\n", _clkfreq/1000000, (_clkfreq/1000)%1000, _clkfreq%1000 );
	printf( " Sampling frequency: %d Hz\n", _clkfreq / SAMPIOD );
	emit_adc_mode( ADCMODE );

	idx = _coginit( COGEXEC_NEW, &startADC, mailbox );	// start another cog to record eight ADC channels via Sinc3 smartpin mode
	if( idx == -1 )  return( 2 );

	// if ADC pins are driven then produce a four cycle square wave during recording
	if( ADCMODE & P_OE )
	{
		idx = SAMPIOD * (SAMPLES>>3) - 4;

__asm {
		dirh	#ADCPINS		// DIR masks the OUT in each cog separately

		rep	@.rend, #8		// four waves
		waitx	idx
		outnot	#ADCPINS
.rend
		fltl	#ADCPINS
}
	}

	_waitatn();				// twiddle thumbs waiting for the other cog to complete

	// emit all the recorded data
	idx = 80;				// start from 10 samples into the recorded data
	printf( "\n P%d P%d P%d P%d P%d P%d P%d P%d\n", ADCPIN, ADCPIN+1, ADCPIN+2, ADCPIN+3, ADCPIN+4, ADCPIN+5, ADCPIN+6, ADCPIN+7 );
	do {
		printf( " %d %d %d %d %d %d %d %d\n", mailbox[idx], mailbox[idx+1], mailbox[idx+2], mailbox[idx+3], mailbox[idx+4], mailbox[idx+5], mailbox[idx+6], mailbox[idx+7] );
		idx += 8;
	} while( idx < 80_000 );

	puts( " STOP" );
	_clkset( 0, 25_000_000 );
	_setbaud( 230400 );

	return( 0 );
}
